import aiohttp
import asyncio
import datetime
import time
import typing
from multidict import CIMultiDict
from .objects import *

class TDSBConnects:

    API_URL = "https://zappsmaprd.tdsb.on.ca/"

    # Platform Name | Client App UUID | Current Time | First App Launch | Version Number | First Launch Current Version | Build Number | First Launch Current Build
    # The server will send a 401 if the build number isn't high enough, however it doesn't check if the number is too high
    # So 2147483647 should ensure that the lib always works
    # The server also doesn't verify any of the fields beside the build num
    X_CLIENT_APP_INFO = "pytdsbconnects||||0.0.0||2147483647|"

    def __init__(self, auto_refresh: bool=True, min_token_life: float=30.0):
        """
        Constructor.

        Although the session is created when the object is constructed, an access token
        is not obtained. login() must be called before any operations.
        
        :param auto_refresh: Whether to automatically refresh the token (default true).
        :param min_token_life: The minimum remaining time before expiry for a token
                               before it is refreshed (in seconds, default 30).
        """
        self._auto_refresh = auto_refresh
        self._min_token_life = min_token_life
        self._token = None
        self._refresh_token = None
        self._token_expiry = None
        self._session = aiohttp.ClientSession(raise_for_status=True, headers={
            "X-Client-App-Info": self.X_CLIENT_APP_INFO,
        })
    
    async def close(self) -> None:
        """
        Close the session.
        """
        await self._session.close()
        await asyncio.sleep(0.250)
    
    async def __aenter__(self) -> "TDSBConnects":
        return self
    
    async def __aexit__(self, exc, *exc_info):
        return await self.close()

    def _update_auth(self, data: typing.Dict[str, typing.Any]) -> None:
        """
        Updates the auth token, refresh token and expiry time.

        For internal use only.

        :param data: A dict containing the auth data.
        """
        self._token = data["access_token"]
        self._refresh_token = data["refresh_token"]
        self._token_expiry = time.time() + data["expires_in"]
        self._session._default_headers = CIMultiDict({
            "Authorization": "Bearer " + self._token,
            "X-Client-App-Info": self.X_CLIENT_APP_INFO,
        })
    
    async def login(self, username: str, password: str) -> None:
        """
        Login to the system. This must be called before any operations.

        :param username: The username (student ID).
        :param password: The password.
        """
        resp = await self._session.post(self.API_URL + "token", data={
            "username": username,
            "password": password,
            "grant_type": "password",
        })
        self._update_auth(await resp.json())
    
    async def refresh_token(self) -> None:
        """
        Refresh the access token.
        """
        if self._refresh_token is None:
            return
        resp = await self._session.post(self.API_URL + "token", data={
            "refresh_token": self.refresh_token,
            "grant_type": "refresh_token",
        })
        self._update_auth(await resp.json())
    
    async def refresh_if_expired(self) -> None:
        """
        Refresh the access token if it has expired.
        """
        if self._refresh_token is None:
            return
        if time.time() + self._min_token_life >= self._token_expiry:
            await self.refresh_token()
    
    async def _get_endpoint(self, endpoint) -> typing.Dict[str, typing.Any]:
        """
        Make a GET request to an endpoint.

        For internal use only.

        :param endpoint: The endpoint.
        """
        if self._token is None:
            return
        if self._auto_refresh:
            await self.refresh_if_expired()
        resp = await self._session.get(self.API_URL + endpoint)
        data = await resp.json()
        return data
    
    async def get_timetable(self, school_id: int, date: typing.Union[datetime.date, datetime.datetime]):
        url = f"api/TimeTable/GetTimeTable/Student/{school_id}/{date.day:02d}{date.month:02d}{date.year}"
        return await self._get_endpoint(url)
    
    async def get_user_info(self) -> User:
        return User(self, await self._get_endpoint("api/Account/GetUserInfo"))



import datetime
import enum
import typing
import re
from .util import *


class Role(enum.Enum):
    """
    Represents a user's role in a school. Multiple roles may exist.
    """
    UNAUTHORIZED = 0
    ADMIN = 1
    STAFF = 2
    STUDENT = 3
    SUPERINTENDENT = 4
    PRINCIPAL = 5
    VICE_PRINCIPAL = 6
    TEACHER = 7
    EXEC_SUPERINTENDENT = 8


class APIObject:
    """
    Base class for all API objects.
    """
    
    def __init__(self, session: "TDSBConnects", data: typing.Dict[str, typing.Any]):
        self._session = session
        self._data = data


class TimetableItem(APIObject):
    """
    An item in a (student's) timetable.
    """

    @property
    def student_number(self) -> str:
        return self._data["StudentNumber"]
    
    @property
    def course_key(self) -> str:
        return self._data["CourseKey"]
    
    @property
    def course_code(self) -> str:
        return self._data["StudentCourse"]["ClassCode"]
    
    @property
    def course_period(self) -> str:
        return self._data["StudentCourse"]["Period"]
    
    @property
    def course_block(self) -> str:
        return self._data["StudentCourse"]["Block"]
    
    @property
    def course_teacher_name(self) -> str:
        return self._data["StudentCourse"]["TeacherName"]
    
    @property
    def course_room(self) -> str:
        return self._data["StudentCourse"]["RoomNo"]
    
    @property
    def course_school_code(self) -> str:
        return self._data["StudentCourse"]["SchoolCode"]
    
    @property
    def course_date(self) -> datetime.datetime:
        return parse_datetime(self._data["StudentCourse"]["Date"])
    
    @property
    def course_cycle_day(self) -> int:
        return self._data["StudentCourse"]["CycleDay"]
    
    @property
    def course_start(self) -> datetime.datetime:
        return parse_datetime(self._data["StudentCourse"]["StartTime"])
    
    @property
    def course_end(self) -> datetime.datetime:
        return parse_datetime(self._data["StudentCourse"]["EndTime"])
    
    @property
    def course_name(self) -> str:
        return self._data["StudentCourse"]["ClassName"]
    
    @property
    def course_teacher_email(self) -> str:
        return self._data["StudentCourse"]["TeacherEmail"]
    
    @property
    def course_semester(self) -> int:
        return self._data["StudentCourse"]["Semester"]
    
    @property
    def course_term(self) -> int:
        return self._data["StudentCourse"]["Term"]
    
    @property
    def course_timeline(self) -> str:
        return self._data["StudentCourse"]["Timeline"]
    
    @property
    def course_track(self) -> str:
        return self._data["StudentCourse"]["SchoolYearTrack"]


class School(APIObject):
    """
    A school.
    """

    DAY_CYCLE_REGEX = re.compile(r"\w{3}\(([D\d]*)\)")

    @property
    def name(self) -> str:
        return self._data["SchoolName"]
    
    @property
    def code(self) -> int:
        return self._data["SchoolCode"]
    
    @property
    def is_onboard(self) -> str:
        return self._data["IsOnboard"]
    
    @property
    def id(self) -> int:
        """
        The ID of the school *for this user* (NOT an absolute ID; for that use the school code).
        """
        return self._data["SchoolSetting"]["Id"]
    
    @property
    def school_year(self) -> str:
        return self._data["SchoolSetting"]["CurrentSession"]
    
    @property
    def track(self) -> str:
        return self._data["SchoolSetting"]["SchoolYearTrack"]
    
    @property
    def school_year_start(self) -> datetime.datetime:
        return parse_datetime(self._data["SchoolSetting"]["SessionStart"])
    
    @property
    def school_year_end(self) -> datetime.datetime:
        return parse_datetime(self._data["SchoolSetting"]["SessionEnd"])
    
    async def timetable(self, date: datetime.date) -> typing.List[TimetableItem]:
        """
        Get the user's timetable for this school for a date as a list of timetable items.

        :param date: The date to get the timetable for (a datetime or a date).
        """
        url = f"api/TimeTable/GetTimeTable/Student/{self.code}/{date.day:02d}{date.month:02d}{date.year}"
        data = await self._session._get_endpoint(url)
        return [TimetableItem(self._session, item) for item in data["CourseTable"]]

    async def day_cycle_names(self, start_date: datetime.date, end_date: datetime.date):
        """
        Get the cycle names for a range of dates, usually observed as <weekday>(D1).

        :param start_date: The start of the range of dates to query (the first entry in the returned list corresponds to this date)
        :param end_date: The end of the range of dates to query (the last entry in the returned list corresponds to this date)
        """

        url = f"api/TimeTable/GetDayNameDayCycle/{self.code}/{self.school_year}/{self.track}/{start_date.day:02d}{start_date.month:02d}{start_date.year}/{end_date.day:02d}{end_date.month:02d}{end_date.year}"
        data = await self._session._get_endpoint(url)
        return [School.DAY_CYCLE_REGEX.match(x).group(1) for x in data]


class User(APIObject):
    """
    A user.
    """

    @property
    def email(self) -> str:
        return self._data["Email"]
    
    @property
    def id(self) -> str:
        return self._data["UserId"]
    
    @property
    def name(self) -> str:
        return self._data["UserName"]
    
    @property
    def gender(self) -> str:
        return self._data["Gender"]
    
    @property
    def age(self) -> str:
        return self._data["Age"]
    
    @property
    def aw_user_id(self) -> str:
        return self._data["AWUserId"]
    
    @property
    def firstname(self) -> str:
        return self._data["FirstName"]

    @property
    def lastname(self) -> str:
        return self._data["LastName"]
    
    @property
    def picture(self) -> str:
        return self._data["Picture"]
    
    @property
    def thumbnail(self) -> str:
        return self._data["Thumbnail"]
    
    @property
    def principal_emails(self) -> typing.List[str]:
        return self._data["PrincipalEmailsList"]
    
    @property
    def vice_principal_emails(self) -> typing.List[str]:
        return self._data["VicePrincipalEmailsList"]
    
    @property
    def superintendent_emails(self) -> typing.List[str]:
        return self._data["SuperintendentEmailsList"]
    
    @property
    def roles(self) -> typing.List[Role]:
        return [Role(r) for r in self._data["Role"]]
    
    @property
    def birthdate(self) -> datetime.datetime:
        return parse_datetime(self._data["BirthDate"])
    
    @property
    def schools(self) -> typing.List[School]:
        return [School(self._session, school) for school in self._data["SchoolList"]]


from .tdsbconnects import * # nopep8



import datetime
import pytz

def parse_datetime(timestamp: str) -> datetime.datetime:
    """
    Convert a datetime string from the TDSB API to a Python datetime.

    :param timestamp: The timestamp as returned by the API.
    """
    if timestamp.endswith("Z"):
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=datetime.timezone.utc)
    try:
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=pytz.timezone("America/Toronto"))
    except ValueError:
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S%z")



import setuptools
import importlib.util
import os

def get_lib_ver():
    spec = importlib.util.spec_from_file_location("tdsbconnects.version", os.path.join(os.path.dirname(__file__), "tdsbconnects/version.py"))
    version = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(version)
    return version.__version__

lib_ver = get_lib_ver()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytdsbconnects",
    version=lib_ver,
    author="Tyler Tian",
    author_email="tylertian123@gmail.com",
    description="A basic Python TDSB Connects API using aiohttp.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tylertian123/pytdsbconnects",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    install_requires=[
        "pytz",
        "aiohttp"
    ],
    python_requires=">=3.6",
    keywords="tdsb",
    license="MIT",
    project_urls={
        "Bug Tracker": "https://github.com/tylertian123/pytdsbconnects/issues",
        "Source Code": "https://github.com/tylertian123/pytdsbconnects"
    }
)
